Attribute Macro stylus_proc::entrypoint

source ·
#[entrypoint]
For structs, this macro generates a richly-typed entrypoint that parses incoming calldata. For functions, this macro generates a simple, untyped entrypoint that’s bytes-in, bytes-out.

Reentrancy is disabled by default, which will cause the program to revert in cases of nested calls. This behavior can be overridden by passing #[entrypoint(allow_reentrancy = true)]